// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Copyright 2017 and onwards Google, Inc.

#ifndef BAUMWELCH_TRAIN_H_
#define BAUMWELCH_TRAIN_H_

#include <type_traits>
#include <utility>
#include <vector>

#include <fst/types.h>
#include <fst/log.h>
#include <fst/extensions/far/far.h>
#include <fst/compose.h>
#include <fst/fst.h>
#include <fst/shortest-distance.h>
#include <baumwelch/cascade.h>
#include <baumwelch/expectation-table.h>
#include <baumwelch/util.h>

namespace fst {

// Some defaults.
constexpr float kAlpha = 1.;
constexpr int32 kMaxIters = 50;

// Helper for all the options.
struct TrainBaumWelchOptions {
  explicit TrainBaumWelchOptions(int32 max_iters = kMaxIters,
                                 float alpha = kAlpha, float delta = kDelta,
                                 const CascadeOptions &copts = CascadeOptions())
      : max_iters(max_iters), alpha(alpha), delta(delta), copts(copts) {}

  // Maximum number of iterations to perform.
  int32 max_iters;
  // Learning rate.
  float alpha;
  // Comparison/quantization delta used to determine convergence.
  float delta;
  // Options passed to the trainer.
  CascadeOptions copts;
};

namespace internal {

// Computes alpha and beta for idempotent semirings, using A* search during the
// alpha computation. If a state is not visited during search the estimate is
// taken to be semiring zero. This estimate of alpha for a state has the true
// value as an upper bound, since some states not visited during the search will
// have true non-zero values because search terminates once the shortest path
// is found (due to the first_path=true).
template <class Arc, typename std::enable_if<IsIdempotent<
                         typename Arc::Weight>::value>::type * = nullptr>
void ComputeAlphaAndBeta(const ComposeFst<Arc> &ico,
                         std::vector<typename Arc::Weight> *alpha,
                         std::vector<typename Arc::Weight> *beta) {
  // Computes beta.
  ShortestDistance(ico, beta, /*reverse=*/true);
  // Computes alpha using an A* approximation.
  using StateId = typename Arc::StateId;
  using Weight = typename Arc::Weight;
  using MyEstimate = NaturalAStarEstimate<StateId, Weight>;
  using MyQueue = NaturalAStarQueue<StateId, Weight, MyEstimate>;
  using MyArcFilter = AnyArcFilter<Arc>;
  using MyShortestDistanceOptions =
      ShortestDistanceOptions<Arc, MyQueue, MyArcFilter>;
  const MyEstimate estimate(*beta);
  MyQueue queue(*alpha, estimate);
  const MyArcFilter arc_filter;
  const MyShortestDistanceOptions opts(
      &queue, arc_filter,
      /*source=*/kNoStateId,     // Default.
      /*delta=*/kShortestDelta,  // Default.
      /*first_path=*/true);      // Heuristic is admissible.
  ShortestDistance(ico, alpha, opts);
  VLOG(1) << ExploredStates<Weight>(*alpha) << " alpha states explored";
}

// Computes alpha and beta for non-idempotent semirings, which requires full
// search.
template <class Arc, typename std::enable_if<!IsIdempotent<
                         typename Arc::Weight>::value>::type * = nullptr>
void ComputeAlphaAndBeta(const ComposeFst<Arc> &ico,
                         std::vector<typename Arc::Weight> *alpha,
                         std::vector<typename Arc::Weight> *beta) {
  // Computes beta.
  ShortestDistance(ico, beta, /*reverse=*/true);
  // Computes alpha.
  ShortestDistance(ico, alpha, /*reverse=*/false);
}

// Class storing forward and backwards weights
template <class Arc>
class ForwardBackward {
 public:
  using StateId = typename Arc::StateId;
  using Weight = typename Arc::Weight;

  explicit ForwardBackward(const ComposeFst<Arc> &ico) {
    ComputeAlphaAndBeta(ico, &alpha_, &beta_);
  }

  const Weight &Alpha(StateId s) const {
    return ForwardBackward::WeightOrZero(s, alpha_);
  }

  const Weight &Beta(StateId s) const {
    return ForwardBackward::WeightOrZero(s, beta_);
  }

 private:
  static constexpr Weight kZero = Weight::Zero();

  // Returns the shortest distance weight, or semiring zero if the state was
  // not visited during the respective shortest distance computation.
  static const Weight &WeightOrZero(StateId s,
                                    const std::vector<Weight> &weights) {
    return (s < weights.size()) ? weights[s] : kZero;
  }

  std::vector<Weight> alpha_;
  std::vector<Weight> beta_;
};

template <class Arc>
constexpr typename ForwardBackward<Arc>::Weight ForwardBackward<Arc>::kZero;

// Object which holds all necessary information for stepwise or minibatch
// training. It stores the (initial) learning rate and the step counter. For
// more information, see the "sEM" pseudocode (p. 613) in:
//
// Liang, P., and Klein, D. 2009. Online EM for unsupervised models. In
// NAACL, pages 611-619.
template <class Arc, class ExpectationTable>
class StepwiseBaumWelchTrainer {
 public:
  using Cascade = typename ExpectationTable::Cascade;
  using Weight = typename Arc::Weight;

  // Valid values of alpha are usually between [.5, 1.0].
  explicit StepwiseBaumWelchTrainer(
      float alpha = kAlpha, const CascadeOptions &opts = CascadeOptions())
      : alpha_(alpha), opts_(opts), step_(0) {}

  // The Step methods all perform a single step or minibatch of stepwise
  // training.

  // Performs a step of training using the specified "ICO" and expectation
  // table. It builds the expectation table, computes the expectations and
  // likelihood by calling Forward, maximizes the model by calling Backward,
  // increments the step counter, and then returns the likelihood. A likelihood
  // of semiring Zero is returned in the case of composition failure.
  Weight Step(const Fst<Arc> &input, const Fst<Arc> &output,
              MutableFst<Arc> *channel) {
    ExpectationTable table(*channel);
    const auto likelihood = Forward(input, output, *channel, &table);
    if (likelihood == Weight::Zero()) return Weight::Zero();
    Backward(table, channel);
    ++step_;
    return likelihood;
  }

  // TODO(kbg): Add minibatch methods here. These can just take vectors.

  // The Train methods repeatedly call Step methods and are usually used to
  // perform a single pass over the data.

  // Repeatedly do the stepwise computation; pair formulation.
  Weight Train(FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
               MutableFst<Arc> *channel) {
    Summation<Weight> likelihood;  // Tracks iteration likelihood.
    while (!(plaintext->Done() || ciphertext->Done())) {
      likelihood.Add(
          Step(*plaintext->GetFst(), *ciphertext->GetFst(), channel));
      plaintext->Next();
      ciphertext->Next();
    }
    return likelihood.Get();
  }

  // Repeatedly do the stepwise computation; decipherment formulation.
  Weight Train(const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
               MutableFst<Arc> *channel) {
    Summation<Weight> likelihood;  // Tracks iteration likelihood.
    for (; !ciphertext->Done(); ciphertext->Next()) {
      likelihood.Add(Step(plaintext, *ciphertext->GetFst(), channel));
    }
    return likelihood.Get();
  }

 private:
  Weight Forward(const Fst<Arc> &input, const Fst<Arc> &output,
                 const Fst<Arc> &channel, ExpectationTable *table) {
    const Cascade cascade(input, output, channel, opts_);
    const auto &ico = cascade.GetFst();
    const auto start = ico.Start();
    if (start == kNoStateId) {
      VLOG(1) << "Empty lattice";
      return false;
    }
    const internal::ForwardBackward<Arc> fb(ico);
    const auto &likelihood = fb.Beta(start);
    if (likelihood == Weight::Zero()) {
      VLOG(1) << "Start state not coaccessible";
      return Weight::Zero();
    }
    for (StateIterator<ComposeFst<Arc>> siter(ico); !siter.Done();
         siter.Next()) {
      const auto state = siter.Value();
      // Non-coaccessible source state.
      if (fb.Beta(state) == Weight::Zero()) continue;
      const auto ch_state = cascade.ChannelState(state);
      const auto &alpha = fb.Alpha(state);
      for (ArcIterator<ComposeFst<Arc>> aiter(ico, state); !aiter.Done();
           aiter.Next()) {
        const auto &arc = aiter.Value();
        const auto &beta = fb.Beta(arc.nextstate);
        // Non-coaccessible destination state.
        if (beta == Weight::Zero()) continue;
        // The arc expectation is the product of the current weight, alpha,
        // and beta, divided by the overall observation likelihood.
        const auto value =
            Divide(Times(Times(alpha, arc.weight), beta), likelihood);
        const auto ch_nextstate = cascade.ChannelState(arc.nextstate);
        table->Forward(ch_state, arc.ilabel, arc.olabel, value, ch_nextstate);
      }
      const auto weight = ico.Final(state);
      if (weight == Weight::Zero()) continue;
      // The final state expectation is the product of the current weight and
      // alpha, divided by the overall observation likelihood.
      const auto value = Divide(Times(alpha, weight), likelihood);
      table->Forward(ch_state, value);
    }
    return likelihood;
  }

  static Weight Interpolate(const Weight &old_weight, const Weight &new_weight,
                            double nu_k) {
    const auto old_term = Times(1 - nu_k, old_weight);
    const auto new_term = Times(nu_k, new_weight);
    return Summation<Weight>::LogPlus(old_term, new_term);
  }

  void Backward(const ExpectationTable &table, MutableFst<Arc> *channel) {
    const double nu_k = std::pow(step_ + 2, -alpha_);
    for (StateIterator<MutableFst<Arc>> siter(*channel); !siter.Done();
         siter.Next()) {
      const auto state = siter.Value();
      // Sets new arc weights.
      for (MutableArcIterator<MutableFst<Arc>> aiter(channel, state);
           !aiter.Done(); aiter.Next()) {
        auto arc = aiter.Value();
        arc.weight = Interpolate(arc.weight, table.Backward(state, arc), nu_k);
        aiter.SetValue(arc);
      }
      // Sets new final weights.
      channel->SetFinal(state, Interpolate(channel->Final(state),
                                           table.Backward(state), nu_k));
    }
  }

  float alpha_;  // Learning rate parameter.
  const CascadeOptions opts_;
  uint64 step_;  // Iteration/step number.
};

// Simple training setup; pair formulation.
template <class Arc, class ExpectationTable>
typename Arc::Weight TrainBaumWelch(
    FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel,
    const TrainBaumWelchOptions &opts = TrainBaumWelchOptions()) {
  using Weight = typename Arc::Weight;
  auto last_likelihood = Weight::Zero();
  StepwiseBaumWelchTrainer<Arc, ExpectationTable> trainer(opts.alpha,
                                                          opts.copts);
  for (int32 iter = 1; iter <= opts.max_iters; ++iter) {
    plaintext->Reset();
    ciphertext->Reset();
    const auto likelihood = trainer.Train(plaintext, ciphertext, channel);
    LOG(INFO) << "Iteration " << iter << ": " << likelihood;
    if (ApproxEqual(last_likelihood, likelihood, opts.delta)) return likelihood;
    last_likelihood = likelihood;
  }
  return last_likelihood;
}

// Simple training setup; decipherment formulation.
template <class Arc, class ExpectationTable>
typename Arc::Weight TrainBaumWelch(
    const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel,
    const TrainBaumWelchOptions &opts = TrainBaumWelchOptions()) {
  using Weight = typename Arc::Weight;
  auto last_likelihood = Weight::Zero();
  StepwiseBaumWelchTrainer<Arc, ExpectationTable> trainer(opts.alpha,
                                                          opts.copts);
  for (int32 iter = 1; iter <= opts.max_iters; ++iter) {
    ciphertext->Reset();
    const auto likelihood = trainer.Train(plaintext, ciphertext, channel);
    LOG(INFO) << "Iteration " << iter << ": " << likelihood;
    if (ApproxEqual(last_likelihood, likelihood, opts.delta)) return likelihood;
    last_likelihood = likelihood;
  }
  return last_likelihood;
}

}  // namespace internal

// Simple training setup; pair formulation.
template <class Arc>
typename Arc::Weight TrainBaumWelch(
    FarReader<Arc> *plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel, ExpectationTableType etype,
    const TrainBaumWelchOptions &opts = TrainBaumWelchOptions()) {
  using Weight = typename Arc::Weight;
  switch (etype) {
    case GLOBAL: {
      return internal::TrainBaumWelch<Arc, GlobalExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case ILABEL: {
      return internal::TrainBaumWelch<Arc, ILabelExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case STATE: {
      return internal::TrainBaumWelch<Arc, StateExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case STATE_ILABEL: {
      return internal::TrainBaumWelch<Arc, StateILabelExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
  }
  // Unreachable.
  return Weight::Zero();
}

template <class Arc>
typename Arc::Weight TrainBaumWelch(
    const Fst<Arc> &plaintext, FarReader<Arc> *ciphertext,
    MutableFst<Arc> *channel, ExpectationTableType etype,
    const TrainBaumWelchOptions &opts = TrainBaumWelchOptions()) {
  using Weight = typename Arc::Weight;
  switch (etype) {
    case GLOBAL: {
      return internal::TrainBaumWelch<Arc, GlobalExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case ILABEL: {
      return internal::TrainBaumWelch<Arc, ILabelExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case STATE: {
      return internal::TrainBaumWelch<Arc, StateExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
    case STATE_ILABEL: {
      return internal::TrainBaumWelch<Arc, StateILabelExpectationTable<Arc>>(
          plaintext, ciphertext, channel, opts);
    }
  }
  // Unreachable.
  return Weight::Zero();
}

}  // namespace fst

#endif  // BAUMWELCH_TRAIN_H_

